#ifndef DEBUG
#define DEBUG (1)
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/file.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/mman.h>
#include <errno.h>
/*#include <pcimem.h>*/
#include "memmap.h"

/* local var(s) */

static char *devname[] =
{
    "/dev/"DEVNAME"0",
    "/dev/"DEVNAME"1",
    "/dev/"DEVNAME"2",
    "/dev/"DEVNAME"3",
    "/dev/"DEVNAME"4",
    "/dev/"DEVNAME"5",
    "/dev/"DEVNAME"6",
    "/dev/"DEVNAME"7",
};

static int mapped[][2] =
{
    {0, 0},
    {0, 0},
    {0, 0},
    {0, 0},
    {0, 0},
    {0, 0},
    {0, 0},
    {0, 0},
};

static int fd[] =
{
    -1, -1, -1, -1, -1, -1, -1, -1,
};

static unsigned int * rptr[NPCIMEM];
static unsigned int * mptr[NPCIMEM];

/* local func(s) */
static void TBmemMap(int devid);

#ifdef __linux__
void
TBgetDmaInfo(int devid, unsigned long *pa, unsigned int **va)
     /*
      * pa: physical addr of DMA buf alloced in the kernel space
      * va: virtual addr (in the user process context) pa is mapped to
      */
{
    size_t psz = getpagesize();
    unsigned long tmp;

    if (-1 == ioctl(fd[devid], GET_DMA_INFO, pa))
    {
	perror("TBgetDmaInfo() failed");
	exit (1);
    };
    if (-1 == ioctl(fd[devid], SET_MAP_MODE, MAP_DMABUF))
    {
	perror("TBgetDmaInfo() failed");
	exit (1);
    };
    fprintf(stderr,"Try mmap\n");
    *va = mmap(NULL,
	       ((PHIBRAMSIZE-1)/psz+1)*psz,
	       PROT_READ | PROT_WRITE,
	       MAP_SHARED,
	       fd[devid],
	       0);
    fprintf(stderr,"return mmap\n");
    if (-1 == ioctl(fd[devid], SET_MAP_MODE, MAP_PHIBREG))
    {
	perror("TBgetDmaInfo() failed");
	exit (1);
    };
    if ((long int)(*va) == -1)
    {
	perror("TBgetDmaInfo");
	exit (1);
    }
    fprintf(stderr,"return dmainfo\n");
}

unsigned int touch_page_dummy = 1;
void
touch_page(unsigned int *ptr, int size)
{
    int i;
    *ptr = touch_page_dummy;

    for (i = 0; i < size; i++)
    {
	ptr[i] = i;
    }
}


unsigned long
TBdmaMapLoad(int devid, unsigned int *buf, int size)
{
    unsigned long pa;
    int psz = getpagesize();
    int off;

    touch_page(buf, size);

    pa = (unsigned long)buf/psz*psz;
    off = (unsigned long)buf - pa;

    /*    fprintf(stderr, "virtual: buf: %08x pa: %08x off: %08x\n",
	    buf, pa, off);*/

    if (-1 == ioctl(fd[devid], DMA_MAP_LOAD, &pa))
    {
	perror("TBdmaMapLoad() failed");
	exit (1);
    };
    (unsigned long)buf = pa + off;

    /*    fprintf(stderr, "physical: buf: %08x pa: %08x\n", buf, pa);*/

    return ((unsigned long)buf);
}

#elif defined(__DECC)
#if NODMA
unsigned long
TBdmaMapLoad(int devid, unsigned int * buf, int size)
{
    fprintf(stderr, "TBdmaMapLoad() not implemented yet.\n");
}

#else /* do DMA */


unsigned long
TBdmaMapLoad(int devid, unsigned int *buf, int size)
{
    unsigned long pa;
    int psz = getpagesize();
    int off;

    touch_page(buf, size);

    pa = (unsigned long)buf/psz*psz;
    off = (unsigned long)buf - pa;
    /*
    fprintf(stderr, "virtual: buf: %08x pa: %08x off: %08x\n",
	    buf, pa, off);
    */
    if (-1 == ioctl(fd[devid], DMA_MAP_LOAD, &pa))
    {
	perror("TBdmaMapLoad() failed");
	exit (1);
    };
    (unsigned long)buf = pa + off;
    /*
    fprintf(stderr, "physical: buf: %08x pa: %08x\n", buf, pa);
    */
    return ((unsigned long)buf);
}
#endif /* NODMA */

#else
#error "unsupported environment"
#endif

unsigned int
TBconfigRead(int devid, unsigned int addr)
{
    struct long_access ca;
    ca.addr = addr;
    ca.data = 0;

    if (fd[devid] < 0)
    {
	fprintf(stderr, "open %s first\n", devname[devid]);
	exit (1);
    }
    ioctl(fd[devid], READ_CFG, &ca);
    return (ca.data);
}

void
TBconfigWrite(int devid, unsigned long addr, unsigned int value)
{
    struct long_access ca;
    ca.addr = addr;
    ca.data = value;

    if (fd[devid] < 0)
    {
	fprintf(stderr, "open %s first\n", devname[devid]);
	exit (1);
    }
    ioctl(fd[devid], WRITE_CFG, &ca);
}

void
TBregWrite(int devid, unsigned long addr, unsigned int value)
{
    rptr[devid][addr>>2] = value;
    MB;
}

unsigned int
TBregRead(int devid, unsigned long addr)
{
    unsigned int ret;

    ret =  rptr[devid][addr>>2];
    return (ret);
}

void
TBmemWrite(int devid, unsigned long addr, unsigned int value)
{
    mptr[devid][addr] = value;
    MB;
}

unsigned int
TBmemRead(int devid, unsigned long addr)
{
    unsigned int ret;

    ret =  mptr[devid][addr];
    MB;
    return (ret);
}


#if defined(__linux__)

static void
TBmemMap(int devid)
{
    unsigned int bar2;
    size_t psz = getpagesize();

    if (mapped[devid][1]) return;
    if (fd[devid] < 0)
    {
	fprintf(stderr, "open %dth device first\n",
		devid);
	exit (1);
    }
    bar2 = TBconfigRead(devid, 0x18);
    bar2 &= 0xfffffff0;
    /*
    printf("bar2: 0x%08x\n", bar2);
    */
    mptr[devid] = mmap(NULL,
		       ((MEM_SIZE+REG_SIZE)/psz+1)*psz,
		       PROT_READ | PROT_WRITE,
		       MAP_SHARED,
		       fd[devid],
		       0); /* device driver (pcimem) take care of bar2 offset */

    if ((long int)mptr[devid] == -1)
    {
	fprintf(stderr, "%s\n", devname[devid]);
	exit (1);
    }
    mptr[devid] += bar2 % psz;
    rptr[devid] = mptr[devid]+(MEM_SIZE>>2);

    TBconfigWrite(devid, 0x10, bar2+MEM_SIZE);

#if 0
    printf("mptr: 0x%lx\n", mptr[devid]);
    printf("rptr: 0x%lx\n", rptr[devid]);
#endif
    mapped[devid][1] = 1;
}

#elif defined (__DECC)

void
TBmemMap(int devid)
{
  unsigned long bar2;

  if (mapped[devid][1]) return;
  if (fd[devid] < 0)
    {
      fprintf(stderr, "open %dth device first\n",
	      devid);
      exit (1);
    }

  if (ioctl(fd[devid], MAP_BAR2_DENSE_ADDR) < 0)
    {
      fprintf(stderr, "ioctl failed for %dth device\n",
	      devid);
      exit (1);
    }

  mptr[devid] = mmap(NULL,
		     MEM_SIZE+REG_SIZE,
		     PROT_WRITE,
		     MAP_FILE|MAP_PRIVATE,
		     fd[devid],
		     0);

  rptr[devid] = (unsigned int *)(((char *)(mptr[devid]))+MEM_SIZE);

  bar2 = TBconfigRead(devid, 0x18);
  TBconfigWrite(devid, 0x10, (bar2 & 0xfffffff0)+MEM_SIZE);

#if 0
  printf("mptr: 0x%lx\n", mptr[devid]);
  printf("rptr: 0x%lx\n", rptr[devid]);
#endif

  if ((int)mptr[devid] == -1)
    {
      fprintf(stderr, "mmap failed for %dth device\n",
	      devid);
      exit (1);
    }
  mapped[devid][1] = 1;

}

#else
#error "unsupported environment"
#endif

/*
  TBopen(int devid)

  return values
  non 0: mapped address
  0: failure
  */
caddr_t
TBopen(int devid)
{
    int ic;
    static int firstcall = 1;
    char buff[256];
    int ifd;
    if (firstcall){
	for (ic = 0; ic < NPCIMEM; ic++){
	    rptr[ic] = NULL;
	    mptr[ic] = NULL;
	}
	firstcall = 0;
    }
    
    /*    fprintf(stderr,"TBopen: Open device %d = %s\n", devid,devname[devid]);*/
    
    fd[devid] = open(devname[devid], O_RDWR);
    if (fd[devid] < 0){
	sprintf(buff, "%s open failed", devname[devid]);
	perror(buff);
	return (0);
    }
    ifd = fd[devid];
    
    
#ifdef LINUX
    if(flock(ifd,LOCK_EX|LOCK_NB)){
	if(errno == EWOULDBLOCK){
	    fprintf(stderr,"Someone is using link, sleep..\n");
	    if(flock(ifd,LOCK_EX)){
		fprintf(stderr,"FLOCK failed with %d\n", errno);
		perror("flock");
		return NULL;
	    }else{
		fprintf(stderr,"Got link...\n");
	    }
	}else{
	    fprintf(stderr,"FLOCK failed with %d\n", errno);
	    perror("flock");
	    return NULL;
	}
    }
#endif
    TBmemMap(devid);
    return (caddr_t) (mptr[devid]);
}

void
TBterm(int devid)
{
#ifdef LINUX
    /*    fprintf(stderr,"TBterm CLOSE\n");*/
    if(flock(fd[devid],LOCK_UN|LOCK_NB)){ 
	fprintf(stderr,"F(UN)LOCK failed with %d\n", errno);
	perror("flock");
    }
#endif
    close(fd[devid]);
    fd[devid] = -1;
}

